/*
Copyright (c) 2004 EarthLink, Inc.
All rights reserved.

FPeer Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* fspd.java
 * 
 * Created on Apr 23, 2004
 *
 * 
*/
package net.earthlink.research.fsp;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.FileNotFoundException;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import java.util.Enumeration;
import java.util.Hashtable;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.ConsoleAppender;

/**
 * @author petrovic
 *
 */
public class fspd extends Thread {

   private boolean         spin;
   private int             serverPort;
   private int             fileNameMode;
   private String          serverAddress;
   private String          directory;
   private DatagramSocket  socket;
   private FSPpacket       fspPacket;
   
   volatile private Hashtable       fileHash       = new Hashtable();
   
   private Logger          logger;
   private PatternLayout   layout;
   private ConsoleAppender appender;
   
   private   final char[]  toHex = {'0','1','2','3','4','5','6','7',
         '8','9','a','b','c','d','e','f'};

   public final static int MODE_DIGEST = 0;   
   public final static int MODE_CLEAR  = 1;
   
   //-------------------------------------------------------------
   
   public fspd( String address, String cacheDirectory) {
      this(address, 2121, cacheDirectory, MODE_CLEAR);            
   }

   public fspd(String address, int port, String dir, int mode ) {
      directory     = dir;
      fileNameMode  = mode;
      serverAddress = address;
      serverPort    = port;
      init(address, port);
   }

   private void init(String address, int port) {
      fspPacket = new FSPpacket();
      spin      = true;

      logger   = Logger.getLogger(this.getClass().getPackage().getName());
      layout   = new PatternLayout("%d{dd MMM yyyy HH:mm:ss}, %l, %m %n");
      appender = new ConsoleAppender(layout);
      logger.addAppender(appender);
      logger.setLevel((Level ) Level.WARN);

      try {
         socket = new DatagramSocket(port, InetAddress.getByName(address));
      } catch (UnknownHostException e) {
         e.printStackTrace();
      } catch (SocketException e) {
         e.printStackTrace();
      } catch (SecurityException e) {
         e.printStackTrace();
      }

      /* 
       * This hash scheme can be improved.
       * Need a directory watcher of some sort to update the hash
       * intelligently. 
       */
      
      if( fileNameMode == MODE_DIGEST ) {
         makeHash();      
         new Thread() {
            public void run() {
               while (spin) {                
                  try {
                     sleep(60000);
                     makeHash();
                  }
                  catch (InterruptedException e) {
                     e.printStackTrace();
                  }
               }
            } 
         }.start();
      }
   }

   public void run( ) {      
      byte[]           buf;      
      byte[]           fileIOBuf;
      int              nRead;
      int              clientPort;
      String           fileName;
      String           handle;
      StringBuffer     sb;
      DatagramPacket   receiveUDP;
      DatagramPacket   sendUDP;      
      RandomAccessFile file;
      InetAddress      clientAddress;     

      sb = new StringBuffer();
      sb.append("fspd server started on ").append(serverAddress).append(":").
         append(Integer.toString(serverPort)).append("/").append(directory);
      logger.warn(sb.toString());      

      buf       = new byte[FSPpacket.MAXSIZE];      
      fileIOBuf = new byte[1024];

      while (spin) {
         try {

            receiveUDP = new DatagramPacket(buf, buf.length);  
            socket.receive(receiveUDP);

            clientAddress = receiveUDP.getAddress();
            clientPort    = receiveUDP.getPort();            
            
            if( fspPacket.disassemble(receiveUDP) ) {

               //logger.debug(fspPacket);
               
               // Need to do better/some key mgmt here...
               fspPacket.bb_key = 0x1960;                             

               switch( fspPacket.cmd ) {

                  case FSPpacket.CC_GET_FILE:
                                                               
                     handle   = new String(fspPacket.buf, 0, fspPacket.bb_len - 1);
                     fileName = getFileName(handle);

                     logger.debug("FSP CC_GET_FILE request for handle = " + handle + "/" + fileName);

                     /* Prepend the directory to the filename */
                     sb       = new StringBuffer();
                     fileName = new String(sb.append(directory).append(fileName));                     
                                           
                     try {
                        file = new RandomAccessFile(fileName, "r");
                        file.seek((long )fspPacket.bb_pos);
                        nRead = file.read(fileIOBuf);
                        file.close();

                        fspPacket.bb_len = (nRead == -1) ? 0 : (short) nRead;
                        fspPacket.setData(fileIOBuf, 0, fspPacket.bb_len, null, 0, 0);
                        sendUDP = new DatagramPacket(buf, buf.length, clientAddress, clientPort);                        
                        fspPacket.assemble(sendUDP);                        
                        socket.send(sendUDP);                        
                     }
                     catch (FileNotFoundException e) {
                        logger.fatal("File not found = " + fileName);                        
                        e.printStackTrace();
                     }
                     catch (EOFException e) {
                        logger.error("EOF reached for file = " + fileName);
                        e.printStackTrace();
                     }
                     catch (IOException e) {
                        logger.fatal("IOException");
                        e.printStackTrace();
                     }

                     break;
                     
                  case FSPpacket.CC_BYE:                     
                     try { 
                        fspPacket.setData(null, 0, 0, null, 0, 0);
                        sendUDP = new DatagramPacket(buf, buf.length, clientAddress, clientPort);                        
                        fspPacket.assemble(sendUDP);
                        socket.send(sendUDP);                        
                     }                                          
                     catch (IOException e) {
                        logger.fatal("Socket IOException");
                        e.printStackTrace();
                     }
                     break;

                  case FSPpacket.CC_GET_DIR:                    

                     // This command implementation is incomplete (no cross-1Kb boundary treatment)

                     File     dir;
                     String[] dirList;

                     handle   = new String(fspPacket.buf, 0, fspPacket.bb_len - 1);
                     dir     = new File(directory + "/" + handle);
                     dirList = dir.list();

                     int bufOffset = 0;
                     for( int i = 0; i < dirList.length; ++i ) {
                        byte[] t = makeDirEntry(dirList[i]);

                        /* Really need to check if room in buf for RDIRENT...
                         * For now, blithely copy.  Works for dir lists < 1024 bytes long.
                        */ 
                        System.arraycopy(t, 0, buf, bufOffset, t.length);
                        bufOffset += t.length;
                     }

                     byte[] t = makeRDEnd();
                     System.arraycopy(t, 0, buf, bufOffset, t.length);
                     bufOffset += t.length;

                     fspPacket.setData(buf, 0, bufOffset, null, 0, 0);
                     sendUDP = new DatagramPacket(buf, buf.length, clientAddress, clientPort);                        
                     fspPacket.assemble(sendUDP);
                     socket.send(sendUDP);                        
                     break;
                     
                  case FSPpacket.CC_STAT: 
                     break;

                  case FSPpacket.CC_GET_PRO:                     
                     break;

                  default: break;
               }

            }
            else {
               // send CC_ERR
               logger.error("Bad FSP packet");
            }

         } catch (IOException e) {
            e.printStackTrace();
            spin = false;
         }
      }
      socket.close();
   }
   
   public void stopServer() {
      spin = false;
   }
   
   private void makeHash() {
      File            f;
      File            dir;
      FileInputStream fis;
      String          fileName;
      String          digest;
      String[]        fileList;
      byte[]          b    = null;
      byte            bid[];

      dir      = new File(directory);
      fileList = dir.list();   
      synchronized (fileHash) {
         fileHash.clear();      
         for( int i=0; i < fileList.length; ++i ) {
            fileName = fileList[i];
            f        = new File(directory + "/" + fileName);                 
            try {      
               fis  = new FileInputStream(f);
               b    = new byte[(int )f.length()];
               fis.read(b);
               bid    = MessageDigest.getInstance("MD5").digest(b);
               digest = toHexString(bid);
               fileHash.put(digest, fileName);      
            } catch (FileNotFoundException e) {
               e.printStackTrace();
            } catch (IOException e) {
               e.printStackTrace();
            } catch (NoSuchAlgorithmException e) {
               e.printStackTrace();
            }                           
         }
      }
   }
   
   private String toHexString(byte[] b) {
      int pos = 0;
      char[] c = new char[b.length * 2];
      for (int i = 0; i < b.length; i++) {
         c[pos++] = toHex[(b[i] >> 4) & 0x0F];
         c[pos++] = toHex[b[i] & 0x0f];
      }
      return new String(c);
   }

   private String getFileName(String handle) {
      String key;
      String fileName;

      if( fileNameMode == MODE_CLEAR ) 
         return handle;

      fileName = null;
      synchronized( fileHash ) {
         for( Enumeration e = fileHash.keys(); e.hasMoreElements(); ) {                     
            key = (String )e.nextElement();
            if ( key.equals(handle) ) { 
               fileName = (String )fileHash.get(key);
               break;
            }
         }
      }
      return fileName;
   }

   private byte[] makeDirEntry(String fileName) {
      byte   mode;
      byte[] r;
      byte[] asciizBuf;
      long   size;
      long   time;
      int    fileNameOffs = 9;
      int    dataLen;
      int    pad;
      File   file;

      file  = new File(fileName);
      size  = file.length();
      time  = file.lastModified();
  
      if ( file.isFile() ) 
         mode = 0x01;
      else 
         mode = 0x02;

      asciizBuf = FSPutil.stringToASCIIZ(fileName);
      dataLen   = fileNameOffs + asciizBuf.length;
      pad       = ( 4 - (dataLen % 4) ) % 4;

      //logger.warn("Filename = " + fileName + ", datalen = " + dataLen + ", pad = " + pad );

      r    = new byte[dataLen + pad];

      r[0] = (byte )((time >>> 24) & 0xff);
      r[1] = (byte )((time >>> 16) & 0xff);
      r[2] = (byte )((time >>>  8) & 0xff);
      r[3] = (byte )( time         & 0xff);

      r[4] = (byte )((size >>> 24) & 0xff);
      r[5] = (byte )((size >>> 16) & 0xff);
      r[6] = (byte )((size >>>  8) & 0xff);
      r[7] = (byte )( size         & 0xff);

      r[8] = mode;

      System.arraycopy(asciizBuf, 0, r, fileNameOffs, asciizBuf.length);

      for( int i = 0; i < pad; ++i )
         r[fileNameOffs + asciizBuf.length + i] = 0x00;

      return r;
   }

   private byte[] makeRDEnd() {
      return new byte[12];
   }
}
